"""
Module realize descriptors mixin
"""

from typing import Awaitable, Optional, Union

from luna3.common.luna_response import LunaResponse

from .base import Backport3MixinStub


class DescriptorMixin(Backport3MixinStub):
    """Descriptor Mixin to work with descriptors."""

    def extractDescriptors(
        self,
        body: Optional[bytes] = None,
        filename: Optional[str] = None,
        contentType: Optional[str] = None,
        warpedImage: bool = False,
        estimateAttributes: bool = False,
        estimateEmotions: bool = False,
        estimateEthnicities: bool = False,
        estimateHeadPose: bool = False,
        estimateQuality: bool = False,
        scoreThreshold: float = 0,
        extractDescriptor: bool = True,
        extractExif: bool = False,
        pitchLt: Optional[float] = None,
        yawLt: Optional[float] = None,
        rollLt: Optional[float] = None,
        noCache: bool = False,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Extract descriptors from a image. Image can be represented as raw bytes or path to file.

        Args:
            body: image's bytes
            filename: path to the folder with the image
            contentType: image mime type, if contentType is not set, will try to determine it by ourselves
                raise ValueError if mimetype of content is not matches with acceptable formats
                (Available mimetypes are: image/jpeg, image/png, image/gif, image/bmp, image/tiff,
                application/x-vl-xpk, application/x-vl-face-descriptor, image/x-portable-pixmap)
            warpedImage: Determines, whether an input image is a warped or an arbitrary one. Exact image warping
                algorithm is proprietary and this flag is intended for VisionLabs front-end tools only.

                The warped image has the following properties:

                * size is always 250x250 pixels;

                * color format is always RGB;

                * single face in a photo;

                * the face is always centered and rotated so that imaginary line between the eyes is horizontal.
            estimateAttributes: whether to estimate face attributes from the image or not(gender, age, glasses).
            estimateEmotions: Whether to estimate emotions from the image.
            estimateEthnicities: Whether to estimate ethnicities from the image.
            estimateQuality: whether to estimate faces' suitability for recognition or not
            scoreThreshold: If estimate_quality parameter is set to 1, it is possible to apply a threshold check
                 to each estimation. All face detections with quality below the threshold are
                 ignored and no descriptors are extracted from them. The function proceeds as
                 usual with all the remaining detections (if left).
            extractDescriptor: whether to extract face descriptor(s) or nor.
            estimateHeadPose: whether to estimate head pose from the image.
            yawLt: maximum deviation yaw angle from 0 (maximum 180)
            pitchLt: maximum deviation pitch angle from 0 (maximum 180)
            rollLt: maximum deviation roll angle from 0 (maximum 180)
            extractExif: Whether to extract EXIF meta information from the input image or not.

                Exact output varies since there are no mandatory data writing requirements both to the authoring
                software and digital cameras.

                This function parses only the tags and outputs their names and values as-is.
            noCache: A boolean, whether to use cache or no

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        if body is None and filename is None:
            raise ValueError

        if body is None:
            with open(filename, "rb") as imageFile:
                body = imageFile.read()

        if contentType is None:
            contentType = self.getContentType(body)

        headers = {"Content-Type": contentType}

        queries = {
            "estimate_attributes": int(estimateAttributes),
            "estimate_emotions": int(estimateEmotions),
            "estimate_ethnicities": int(estimateEthnicities),
            "estimate_quality": int(estimateQuality),
            "extract_descriptor": int(extractDescriptor),
            "warped_image": int(warpedImage),
            "extract_exif": int(extractExif),
            "score_threshold": scoreThreshold,
            "estimate_head_pose": int(estimateHeadPose),
            "pitch__lt": pitchLt,
            "yaw__lt": yawLt,
            "roll__lt": rollLt,
            "no_cache": int(noCache),
        }

        return self.makeAuthRequest(
            url=f"{self.baseUri}/storage/descriptors",
            method="POST",
            queryParams=queries,
            body=body,
            headers=headers,
            **kwargs,
        )

    def getDescriptors(
        self, page: Optional[int] = 1, pageSize: Optional[int] = 10, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        The function gets accounts' descriptors. Result is a list of descriptors and number of descriptors
        for the account. Every descriptor is represented by *descriptor_id* and number of  *linked_lists*, the
        descriptor is attached to.

        Args:
            page: page number, positive.
            pageSize: number of results per page, positive.

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with descriptors.

        """
        queries = {"page": page, "page_size": pageSize}
        return self.makeAuthRequest(
            url=f"{self.baseUri}/storage/descriptors", method="GET", queryParams=queries, **kwargs
        )

    def getDescriptor(self, descriptorId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        The function gets the descriptor, which corresponds to *descriptorId*.

        Args:
            descriptorId: descriptor *id*

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with descriptor.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/storage/descriptors/{descriptorId}", method="GET", **kwargs)

    def getPortrait(self, descriptorId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        The function gets the portrait, which corresponds to *descriptorId*.

        Args:
            descriptorId: descriptor id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with portrait.

        """
        return self.makeAuthRequest(url=f"{self.baseUri}/storage/portraits/{descriptorId}", method="GET", **kwargs)
