"""
C.O.W. Aiohttp errors processing.

See Also:
    http://git.visionlabs.ru/luna/crutches_on_wheels/blob/dev/errors/errors.py
    https://docs.aiohttp.org/en/stable/client_reference.html#hierarchy-of-exceptions
"""
import asyncio
import traceback
from typing import Dict, NamedTuple, Union

from aiohttp import client_exceptions as ce

from .luna_request import LunaRequest


class _ErrorInfo(NamedTuple):
    """
    Class for ErrorInfo stub from C.O.W.

    References:
        http://git.visionlabs.ru/luna/crutches_on_wheels/blob/dev/errors/errors.py
    """

    #: error code
    errorCode: int
    #: error description
    desc: str
    #: error detail
    detail: str

    def formatAsDict(self, *formatArgs) -> Dict[str, Union[int, str]]:
        """
        Convert error to dict. With formatting detail.

        Returns:
             dict with keys "error_code", "desc", "detail"
        """
        return {"error_code": self.errorCode, "desc": self.desc, "detail": self.detail.format(*formatArgs)}


_DEFAULT_ERROR = _ErrorInfo(1, "Unknown HTTP client error", "{}")

_TIMEOUT_ERRORS_MAP = {
    "sock_read": _ErrorInfo(6, "Socket read timeout", "Request timeout on {}, method {}"),
    "sock_connect": _ErrorInfo(7, "Socket connect timeout", "Request socket connect timeout on {}, method {}"),
    "connect": _ErrorInfo(8, "Connect timeout", "Connect timeout on {}, method {}"),
    "total": _ErrorInfo(9, "Request timeout", "Request timeout on {}, method {}"),
}

_EXCEPTIONS_AND_ERRORS_MAP = {
    ce.InvalidURL: _ErrorInfo(3, "Invalid url", "{}"),
    ce.ClientPayloadError: _ErrorInfo(4, "Client payload error", "{}"),
    ce.ServerFingerprintMismatch: _ErrorInfo(5, "Server fingerprint mismatch", "{}"),
    # <- timeout errors
    ce.ServerDisconnectedError: _ErrorInfo(10, "Server disconnected", "{}"),
    ce.ServerConnectionError: _ErrorInfo(11, "Server connection error", "{}"),
    ce.ClientProxyConnectionError: _ErrorInfo(12, "Client proxy connection error", "{}"),
    ce.ClientConnectorSSLError: _ErrorInfo(13, "Client connector SSL error", "{}"),
    ce.ClientConnectorCertificateError: _ErrorInfo(14, "Client connector certificate error", "{}"),
    ce.ClientSSLError: _ErrorInfo(15, "Client SSL error", "{}"),
    ce.ClientConnectorError: _ErrorInfo(16, "Client connector error", "{}"),
    ce.ClientOSError: _ErrorInfo(17, "Client OS error", "{}"),
    ce.ClientConnectionError: _ErrorInfo(18, "Client connection error", "{}"),
    ce.ClientHttpProxyError: _ErrorInfo(19, "Client HTTP proxy error", "{}"),
    ce.WSServerHandshakeError: _ErrorInfo(20, "WS server handshake error", "{}"),
    ce.ContentTypeError: _ErrorInfo(21, "Content-Type error", "{}"),
    ce.ClientResponseError: _ErrorInfo(22, "Client response error", "{}"),
    ce.ClientError: _ErrorInfo(23, "HTTP client error", "{}"),
}


def guessAiohttpError(exc: Exception, request: LunaRequest) -> Dict[str, Union[int, str]]:
    """
    Guess an error on raised exception during request processing:
        1. check server connect timeouts
        2. check total request timeout
        3. except any aiohttp client exception

    References:
        https://docs.aiohttp.org/en/stable/client_reference.html#hierarchy-of-exceptions

    Args:
        exc: raised exception
        request: sent request

    Returns:
        dict:
            errorCode (int): error code
            desc (str): short error description
            detail (str): full error description
    """
    # 1 find server timeout error
    if isinstance(exc, ce.ServerTimeoutError):
        err = _TIMEOUT_ERRORS_MAP["connect"].formatAsDict(request.url, request.method)

        if str(exc) == "Timeout on reading data from socket":
            err = _TIMEOUT_ERRORS_MAP["sock_read"].formatAsDict(request.url, request.method)
        elif str(exc).startswith("Connection timeout to host"):
            if isinstance(exc.__cause__, asyncio.TimeoutError):
                # guess what happened
                stk = traceback.extract_tb(exc.__cause__.__traceback__)
                functionNames = [x.name for x in stk]
                if "_create_connection" in functionNames:
                    # _create_connection performs dns lookup and setting up a connection
                    err = _TIMEOUT_ERRORS_MAP["sock_connect"].formatAsDict(request.url, request.method)
                else:
                    err = _TIMEOUT_ERRORS_MAP["connect"].formatAsDict(request.url, request.method)

    # 2 find total request timeout error
    elif isinstance(exc, asyncio.TimeoutError):
        # Order matters because ServerTimeoutError is a child of asyncio.TimeoutError
        err = _TIMEOUT_ERRORS_MAP["total"].formatAsDict(request.url, request.method)

    # 3 find one of aiohttp client exceptions
    else:
        errorInfo = _EXCEPTIONS_AND_ERRORS_MAP.get(type(exc), _DEFAULT_ERROR)
        err = errorInfo.formatAsDict(f"{request.method}:{request.url}:{exc}")
    return err
