"""
Luna-events client module
"""
from numbers import Real
from typing import Any, Awaitable, Dict, List, Optional, Union

import aiohttp

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..common.base_api import BaseHTTPAPI
from ..common.http_objs import StatQuery
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from ..common.utils import encodeDescriptor
from .http_objs import CLOTHING_COLOR, EVENT_TARGET, PRIMITIVE_CLOTHING_COLOR, Event


class EventsApi(BaseHTTPAPI):
    """
    Class for request to luna-events.

    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5040",
        api: Optional[int] = 2,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def saveEvents(
        self,
        accountId: str,
        events: List[Event],
        waitEventsSaving: Optional[bool] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Save events to luna-events.

        Args:
            accountId: events creator account id
            events: list of events
            waitEventsSaving: whether to wait events saving (response will be received only after events will be saved)
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.

        """
        queries = {"wait_events_saving": waitEventsSaving}

        eventsToSave = []
        if contentType == "application/msgpack":
            builder = RequestPayload.buildMsgpack
            encodeToBase64 = False
        elif contentType == "application/json":
            builder = RequestPayload.buildJson
            encodeToBase64 = True
        else:
            raise ValueError(f"Unsupported Content-Type '{contentType}'")

        for event in events:
            jsonEvent = event if isinstance(event, dict) else event.asDict()
            if encodeToBase64:
                for attrType in ("face", "body"):
                    descriptorData = jsonEvent.get(f"{attrType}_attributes", {}).get("descriptor_data")
                    if descriptorData:
                        descriptorData["descriptor"] = encodeDescriptor(descriptorData["descriptor"], encode=True)
            eventsToSave.append(jsonEvent)

        payload = builder(body=eventsToSave)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/events", "POST", queryParams=queries, body=payload, **kwargs)

    def getEvents(  # pylint: disable=R0914
        self,
        target: Optional[List[EVENT_TARGET]] = None,
        createTimeGte: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        endTimeGte: Optional[str] = None,
        endTimeLt: Optional[str] = None,
        accountId: Optional[str] = None,
        eventIds: Optional[List[str]] = None,
        eventIdGte: Optional[str] = None,
        eventIdLt: Optional[str] = None,
        handlerIds: Optional[List[str]] = None,
        externalIds: Optional[List[str]] = None,
        topMatchingCandidatesLabel: Optional[str] = None,
        topSimilarObjectIds: Optional[List[str]] = None,
        topSimilarExternalIds: Optional[List[str]] = None,
        topSimilarObjectSimilarityGte: Optional[float] = None,
        topSimilarObjectSimilarityLt: Optional[float] = None,
        ageLt: Optional[int] = None,
        ageGte: Optional[int] = None,
        gender: Optional[Union[int, Literal["null"]]] = None,
        emotions: Optional[List[Union[int, Literal["null"]]]] = None,
        masks: Optional[List[Union[int, Literal["null"]]]] = None,
        ethnicGroups: Optional[Union[List[int], Literal["null"]]] = None,
        liveness: Optional[Union[List[int], Literal["null"]]] = None,
        apparentAgeGte: Optional[int] = None,
        apparentAgeLt: Optional[int] = None,
        apparentGender: Optional[List[Union[Literal[0, 1, 2], Literal["null"]]]] = None,
        headwearStates: Optional[List[Union[Literal[0, 1, 2], Literal["null"]]]] = None,
        headwearApparentColors: Optional[List[Union[PRIMITIVE_CLOTHING_COLOR, Literal["null"]]]] = None,
        sleeveLengths: Optional[List[Union[Literal["short", "long", "undefined"], Literal["null"]]]] = None,
        upperClothingColors: Optional[List[Union[CLOTHING_COLOR, Literal["null"]]]] = None,
        lowerGarmentTypes: Optional[
            List[Union[Literal["undefined", "trousers", "shorts", "skirt"], Literal["null"]]]
        ] = None,
        lowerGarmentColors: Optional[List[Union[CLOTHING_COLOR, Literal["null"]]]] = None,
        shoesApparentColors: Optional[List[Union[PRIMITIVE_CLOTHING_COLOR, Literal["null"]]]] = None,
        backpackStates: Optional[List[Union[Literal[0, 1, 2], Literal["null"]]]] = None,
        faceIds: Optional[List[str]] = None,
        userData: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        order: Optional[str] = "desc",
        sources: Optional[List[Union[str, Literal["null"]]]] = None,
        streamIds: Optional[List[Union[str, Literal["null"]]]] = None,
        tags: Optional[List[str]] = None,
        cities: Optional[List[Union[str, Literal["null"]]]] = None,
        areas: Optional[List[Union[str, Literal["null"]]]] = None,
        districts: Optional[List[Union[str, Literal["null"]]]] = None,
        streets: Optional[List[Union[str, Literal["null"]]]] = None,
        houseNumbers: Optional[List[Union[str, Literal["null"]]]] = None,
        originLongitude: Optional[Union[float, Literal["null"]]] = None,
        originLatitude: Optional[Union[float, Literal["null"]]] = None,
        longitudeDelta: Optional[float] = None,
        latitudeDelta: Optional[float] = None,
        trackIds: Optional[List[Union[str, Literal["null"]]]] = None,
        meta: Optional[dict] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get events by filters.

        Args:
            target: list of events' fields to receive instead of full events
            createTimeGte: lower including create time boundary
            createTimeLt: upper excluding create time boundary
            endTimeGte: lower including end time boundary
            endTimeLt: upper excluding end time boundary
            accountId: event account id
            faceIds: ids of faces linked with an event
            eventIds: event ids
            eventIdGte: event id lower including bound
            eventIdLt: event id upper excluding bound
            handlerIds: handler ids
            externalIds: external ids
            userData: event user data
            sources: list of source
            streamIds: list of stream ids
            tags: users tags
            topMatchingCandidatesLabel: top matching candidates label
            topSimilarObjectIds: top similar object ids
            topSimilarExternalIds: top similar external ids
            topSimilarObjectSimilarityGte: lower including top similarity boundary
            topSimilarObjectSimilarityLt: upper excluding top similarity boundary
            ageLt: upper excluding age boundary
            ageGte: lower including age boundary
            gender: gender of event (0 or 1)
            emotions: dominant emotions in event (one of (1, 2, 3, 4, 5, 6, 7) -
                      matches "anger", "disgust", "fear", "happiness", "neutral", "sadness", "surprise")
            masks: dominant masks in event (one of (1, 2, 3) -
                      matches "missing", "medical_mask", "occluded")
            ethnicGroups: dominant ethnicity of the face ((1, 2, 3, 4) -
                                  means (indian, asian, african_american, caucasian)
            liveness: dominant liveness of the face ((0, 1, 2) - means ("spoof", "real", "unknown"))
            apparentAgeLt: apparent age upper excluding boundary
            apparentAgeGte: apparent age upper including boundary
            apparentGender: apparent gender estimation values (0 - female, 1 - male, 2 - undefined)
            headwearStates: headwear states (0 - absent, 1 - present, 2 - undefined)
            headwearApparentColors: headwear apparent colors
            sleeveLengths: sleeve lengths
            upperClothingColors: upper clothing colors
            lowerGarmentTypes: lower garment types
            lowerGarmentColors: lower garment colors
            shoesApparentColors: shoes apparent colors
            backpackStates: backpack states (0 - absent, 1 - present, 2 - undefined)
            areas: list of areas
            streets: list of streets
            districts: list of districts
            cities: list of cities
            houseNumbers: list of house numbers
            originLongitude: origin longitude in degrees
            originLatitude:  origin latitude in degrees
            longitudeDelta: longitude delta in degrees
            latitudeDelta: latitude delta in degrees
            trackIds: list of track ids
            meta: user-defined event meta's filters in format {meta.<path.to.field>__<operator>:<type>: <value>}
            page: number of result page
            pageSize: count of results in page
            order: result sort order (ask or desc)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {
            "target": target,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
            "end_time__gte": endTimeGte,
            "end_time__lt": endTimeLt,
            "account_id": accountId,
            "event_ids": eventIds,
            "event_id__gte": eventIdGte,
            "event_id__lt": eventIdLt,
            "external_ids": externalIds,
            "top_matching_candidates_label": topMatchingCandidatesLabel,
            "top_similar_object_ids": topSimilarObjectIds,
            "top_similar_external_ids": topSimilarExternalIds,
            "top_similar_object_similarity__gte": topSimilarObjectSimilarityGte,
            "top_similar_object_similarity__lt": topSimilarObjectSimilarityLt,
            "age__lt": ageLt,
            "age__gte": ageGte,
            "gender": gender,
            "emotions": emotions,
            "masks": masks,
            "ethnic_groups": ethnicGroups,
            "liveness": liveness,
            "apparent_gender": apparentGender,
            "apparent_age__gte": apparentAgeGte,
            "apparent_age__lt": apparentAgeLt,
            "headwear_states": headwearStates,
            "headwear_apparent_colors": headwearApparentColors,
            "sleeve_lengths": sleeveLengths,
            "upper_clothing_colors": upperClothingColors,
            "lower_garment_types": lowerGarmentTypes,
            "lower_garment_colors": lowerGarmentColors,
            "shoes_apparent_colors": shoesApparentColors,
            "backpack_states": backpackStates,
            "handler_ids": handlerIds,
            "user_data": userData,
            "tags": tags,
            "face_ids": faceIds,
            "sources": sources,
            "stream_ids": streamIds,
            "page": page,
            "page_size": pageSize,
            "order": order,
            "streets": streets,
            "areas": areas,
            "districts": districts,
            "cities": cities,
            "house_numbers": houseNumbers,
            "origin_longitude": originLongitude,
            "origin_latitude": originLatitude,
            "longitude_delta": longitudeDelta,
            "latitude_delta": latitudeDelta,
            "track_ids": trackIds,
            **(meta or {}),
        }
        return self.makeRequest(f"{self.baseUri}/events", "GET", queryParams=queries, **kwargs)

    def getEventDescriptor(
        self,
        eventId: Optional[str] = None,
        accountId: Optional[str] = None,
        descriptorVersion: Optional[int] = None,
        descriptorType: Optional[str] = None,
        acceptType: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get event descriptor by filters.

        Args:
            eventId: event id
            accountId: event account id
            descriptorVersion: descriptor version
            descriptorType: Kind of the descriptor (body or face)
            acceptType: acceptable content-type of response body

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with event descriptor will be returned.

        """
        headers = {"Accept": acceptType} if acceptType else {}
        queries = {"account_id": accountId, "descriptor_version": descriptorVersion}
        if descriptorType:
            queries["descriptor_type"] = descriptorType
        return self.makeRequest(
            f"{self.baseUri}/events/{eventId}/descriptor", "GET", queryParams=queries, headers=headers, **kwargs
        )

    def checkEvent(
        self, eventId: str, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check event existence by id.

        Args:
            eventId: event id
            accountId: event account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {"account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/events/{eventId}", "HEAD", queryParams=queries, **kwargs)

    def getEvent(
        self, eventId: str, target: Optional[List[EVENT_TARGET]] = None, accountId: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get event by id.

        Args:
            eventId: event id
            target: list of events' fields to receive instead of full events
            accountId: event account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {"target": target, "account_id": accountId}
        return self.makeRequest(f"{self.baseUri}/events/{eventId}", "GET", queryParams=queries, **kwargs)

    def patchEvent(
        self,
        eventId: str,
        eventAttributes: Dict[str, Any],
        accountId: Optional[str] = None,
        contentType: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update event existence by id. You can update face or body attributes passed in json.

        Args:
            eventId: event id
            eventAttributes: event attributes for update
            accountId: event account id
            contentType: content type to use (supported now: "application/json", "application/msgpack" (default))

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {"account_id": accountId}

        if contentType == "application/msgpack":
            builder = RequestPayload.buildMsgpack
            encodeToBase64 = False
        elif contentType == "application/json":
            builder = RequestPayload.buildJson
            encodeToBase64 = True
        else:
            raise ValueError(f"Unsupported Content-Type '{contentType}'")

        if encodeToBase64:
            for attrType in ("face", "body"):
                descriptorData = eventAttributes.get(f"{attrType}_attributes", {}).get("descriptor_data")
                if descriptorData:
                    descriptorData["descriptor"] = encodeDescriptor(descriptorData["descriptor"], encode=True)

        payload = builder(body=eventAttributes)
        return self.makeRequest(
            f"{self.baseUri}/events/{eventId}", "PATCH", queryParams=queries, body=payload, **kwargs
        )

    def getEventsStats(
        self, query: Union[StatQuery, dict] = StatQuery(), **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get events statistics by aggregation (<aggregator> and <groupBy> on <target>) and filters.

        Args:
            query: query object or query dict, contained in StatQuery object

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with event statistics will be returned.

        """
        if not isinstance(query, dict):
            query = query.query
        payload = RequestPayload.buildJson(body=query)
        return self.makeRequest(f"{self.baseUri}/events/stats", "POST", body=payload, **kwargs)

    def deleteEvents(
        self,
        target: Optional[List[str]] = None,
        accountId: Optional[str] = None,
        handlerId: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        insertTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete events by filters.

        Args:
            target: event deletion info targets
            accountId: event account id
            handlerId: event handler id
            createTimeGte: lower including create time boundary
            createTimeLt: upper excluding create time boundary
            insertTimeGte: lower including insert time boundary
            insertTimeLt: upper excluding insert time boundary

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {
            "target": target,
            "account_id": accountId,
            "handler_id": handlerId,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
            "insert_time__gte": insertTimeGte,
            "insert_time__lt": insertTimeLt,
        }
        return self.makeRequest(f"{self.baseUri}/events", "DELETE", queryParams=queries, **kwargs)

    def getEventsMissing(
        self,
        kind: str,
        accountId: Optional[str] = None,
        handlerId: Optional[str] = None,
        insertTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        eventIdGte: Optional[str] = None,
        eventIdLt: Optional[str] = None,
        missingVersion: Optional[int] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        order: Optional[str] = "desc",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get events with missing attributes.

        Args:
            kind: kind of missing attribute
            accountId: event account id
            handlerId: handler id
            insertTimeGte: lower including insert time boundary
            insertTimeLt: upper excluding insert time boundary
            eventIdGte: event id lower including bound
            eventIdLt: event id upper excluding bound
            missingVersion: if looking for missing descriptor, look for descriptor of given version
            page: number of result page
            pageSize: count of results in page
            order: result sort order (ask or desc)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {
            "kind": kind,
            "account_id": accountId,
            "handler_id": handlerId,
            "insert_time__gte": insertTimeGte,
            "insert_time__lt": insertTimeLt,
            "event_id__gte": eventIdGte,
            "event_id__lt": eventIdLt,
            "missing_version": missingVersion,
            "page": page,
            "page_size": pageSize,
            "order": order,
        }
        return self.makeRequest(f"{self.baseUri}/events/attributes/missing", "GET", queryParams=queries, **kwargs)

    def getEventsMissingCount(
        self,
        kind: str,
        accountId: Optional[str] = None,
        handlerId: Optional[str] = None,
        insertTimeGte: Optional[str] = None,
        insertTimeLt: Optional[str] = None,
        eventIdGte: Optional[str] = None,
        eventIdLt: Optional[str] = None,
        missingVersion: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get events with missing attributes.

        Args:
            kind: kind of missing attribute
            accountId: event account id
            handlerId: handler id
            insertTimeGte: lower including insert time boundary
            insertTimeLt: upper excluding insert time boundary
            eventIdGte: event id lower including bound
            eventIdLt: event id upper excluding bound
            missingVersion: if looking for missing descriptor, look for descriptor of given version

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {
            "kind": kind,
            "insert_time__gte": insertTimeGte,
            "insert_time__lt": insertTimeLt,
            "handler_id": handlerId,
            "account_id": accountId,
            "event_id__gte": eventIdGte,
            "event_id__lt": eventIdLt,
            "missing_version": missingVersion,
        }
        return self.makeRequest(f"{self.baseUri}/events/attributes/missing/count", "GET", queryParams=queries, **kwargs)

    def deleteDescriptors(
        self,
        descriptorVersion: int,
        descriptorType: str,
        accountId: Optional[str] = None,
        eventIdGte: Optional[str] = None,
        eventIdLt: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete events descriptors.

        Args:
            accountId: event account id
            eventIdGte: event id lower including bound
            eventIdLt: event id upper excluding bound
            descriptorVersion: descriptor version
            descriptorType: Kind of the descriptor (body or face)

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with events will be returned.

        """
        queries = {
            "account_id": accountId,
            "event_id__gte": eventIdGte,
            "event_id__lt": eventIdLt,
            "version": descriptorVersion,
            "descriptor_type": descriptorType,
        }
        return self.makeRequest(f"{self.baseUri}/events/descriptors", "DELETE", queryParams=queries, **kwargs)
