"""
Module realize faces mixin
"""
#  pylint: disable=missing-docstring, C0301

from typing import Awaitable, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from luna3.accounts.http_objs import TokenPermissions
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.lunavl.mixins.base import MixinStub


class AccountsMixin(MixinStub):
    """Accounts Mixin to work with accounts and tokens."""

    def createAccount(
        self,
        login: str,
        password: str,
        accountType: Literal["advanced_user", "user"],
        description: Optional[str] = None,
        accountId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new account

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createAccount

        Args:
            login: account login
            password: account password
            accountType: account type
            description: account description
            accountId: account id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created account id will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from uuid import uuid4
        >>>
        >>> response = client.createAccount(
        ...     accountId=str(uuid4()), login=f"{uuid4()}@visionlabs.ru", password="pwd", accountType="user"
        ... )
        >>> print(response.statusCode)
        201
        >>> print(response.json) # doctest: +ELLIPSIS
        {'account_id': ..., 'url': ...}

        """
        data = dict(login=login, password=password, account_type=accountType)
        if description is not None:
            data["description"] = description
        payload = RequestPayload.buildJson(body=data)
        payload.applyAccountIdHeader(accountId)
        return self.makeRequest(f"{self.baseUri}/accounts", "POST", body=payload, **kwargs)

    def getAccount(
        self,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getAccount

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with account will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from uuid import uuid4
        >>>
        >>> response = client.getAccount()
        >>> print(response.statusCode)
        200
        >>> print(response.json) # doctest: +ELLIPSIS
        {'account_id': ..., 'login': ..., 'account_type': 'advanced_user', 'description': ..., 'create_time': ..., 'last_update_time': ...}

        """
        return self.makeRequest(f"{self.baseUri}/account", "GET", **kwargs)

    def patchAccount(
        self,
        login: Optional[str] = None,
        password: Optional[str] = None,
        accountType: Optional[Literal["advanced_user", "user"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Patch account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/patchAccount

        Args:
            login: account login
            password: account password
            accountType: account type
            description: account description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` 204 status code will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from uuid import uuid4
        >>>
        >>> response = client.patchAccount(description="new description")
        >>> print(response.statusCode)
        204

        """
        data = {}
        if login is not None:
            data["login"] = login
        if password is not None:
            data["password"] = password
        if accountType is not None:
            data["account_type"] = accountType
        if description is not None:
            data["description"] = description
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/account", "PATCH", body=payload, **kwargs)

    def createToken(
        self,
        tokenPermissions: TokenPermissions,
        expirationTime: Optional[str] = None,
        visibilityArea: Optional[Literal["all", "account"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new token.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createToken

        Args:
            tokenPermissions: token permissions
            expirationTime: token expiration time
            visibilityArea: token visibility area
            description: account description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created token id and token will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.accounts.http_objs import TokenPermissions
        >>> from uuid import uuid4
        >>>
        >>> response = client.createToken(tokenPermissions=TokenPermissions())
        >>> print(response.statusCode)
        201
        >>> print(response.json) # doctest: +ELLIPSIS
        {'token_id': ..., 'token': ..., 'url': ...}

        """
        data = {"permissions": tokenPermissions.asDict(), "expiration_time": expirationTime}
        if description is not None:
            data["description"] = description
        if visibilityArea is not None:
            data["visibility_area"] = visibilityArea
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/tokens", "POST", body=payload, **kwargs)

    def getTokens(
        self,
        page: int = 1,
        pageSize: int = 10,
        accountIdFilter: Optional[str] = None,
        createTimeLt: Optional[str] = None,
        createTimeGte: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get tokens.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getTokens

        Args:
            page: page, default 1
            pageSize: page size, default 10
            accountIdFilter: account id to get tokens
            createTimeLt: upper bound for token create_time
            createTimeGte: lower bound for token create_time

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with tokens will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.accounts.http_objs import TokenPermissions
        >>> from uuid import uuid4
        >>> import json
        >>>
        >>> _ = client.createToken(tokenPermissions=TokenPermissions(), raiseError=True)
        >>> response = client.getTokens(pageSize=1)
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "tokens": [
                {
                    "account_id": ...,
                    "create_time": ...,
                    "description": "",
                    "expiration_time": null,
                    "last_update_time": ...,
                    "permissions": {
                        "attribute": [],
                        "body_sample": [],
                        "emit_events": {
                            "allowed": 1,
                            "black_list": null,
                            "white_list": null
                        },
                        "event": [],
                        "face": [],
                        "face_sample": [],
                        "handler": [],
                        "image": [],
                        "lambdas": [],
                        "list": [],
                        "object": [],
                        "resources": [],
                        "task": [],
                        "token": [],
                        "verifier": []
                    },
                    "token": ...,
                    "token_id": ...,
                    "visibility_area": "all"
                }
            ]
        }

        """
        queries = {
            "page": page,
            "page_size": pageSize,
            "account_id": accountIdFilter,
            "create_time__gte": createTimeGte,
            "create_time__lt": createTimeLt,
        }
        return self.makeRequest(f"{self.baseUri}/tokens", "GET", queryParams=queries, **kwargs)

    def replaceToken(
        self,
        tokenId: str,
        tokenPermissions: TokenPermissions,
        expirationTime: Optional[str] = None,
        visibilityArea: Optional[Literal["all", "account"]] = None,
        description: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Replace token.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/putToken

        Args:
            tokenId: token id
            tokenPermissions: token permissions
            expirationTime: token expiration time
            visibilityArea: token visibility area
            description: account description

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with created token id and token will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.accounts.http_objs import TokenPermissions
        >>> from uuid import uuid4
        >>> import json
        >>>
        >>> tokenId = client.createToken(tokenPermissions=TokenPermissions(), raiseError=True).json["token_id"]
        >>> response = client.replaceToken(tokenId=tokenId, tokenPermissions=TokenPermissions())
        >>> print(response.statusCode)
        201
        >>> print(response.json) # doctest: +ELLIPSIS
        {'token_id': ..., 'token': ..., 'url': ...}

        """
        data = {"permissions": tokenPermissions.asDict(), "expiration_time": expirationTime}
        if description is not None:
            data["description"] = description
        if visibilityArea is not None:
            data["visibility_area"] = visibilityArea
        payload = RequestPayload.buildJson(body=data)

        return self.makeRequest(f"{self.baseUri}/tokens/{tokenId}", "PUT", body=payload, **kwargs)

    def getToken(
        self, tokenId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get token.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getToken

        Args:
            tokenId: token id
            accountIdFilter: account id to get token

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with token will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.accounts.http_objs import TokenPermissions
        >>> from uuid import uuid4
        >>> import json
        >>>
        >>> tokenId = client.createToken(tokenPermissions=TokenPermissions(), raiseError=True).json["token_id"]
        >>> response = client.getToken(tokenId=tokenId)
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "create_time": ...,
            "description": "",
            "expiration_time": null,
            "last_update_time": ...,
            "permissions": {
                "attribute": [],
                "body_sample": [],
                "emit_events": {
                    "allowed": 1,
                    "black_list": null,
                    "white_list": null
                },
                "event": [],
                "face": [],
                "face_sample": [],
                "handler": [],
                "image": [],
                "lambdas": [],
                "list": [],
                "object": [],
                "resources": [],
                "task": [],
                "token": [],
                "verifier": []
            },
            "token": ...,
            "token_id": ...,
            "visibility_area": "all"
        }

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/tokens/{tokenId}", "GET", queryParams=queries, **kwargs)

    def deleteToken(self, tokenId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete token.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/deleteToken

        Args:
            tokenId: token id

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` will return nothing in text field

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from luna3.accounts.http_objs import TokenPermissions
        >>> from uuid import uuid4
        >>> import json
        >>>
        >>> tokenId = client.createToken(tokenPermissions=TokenPermissions(), raiseError=True).json["token_id"]
        >>> response = client.deleteToken(tokenId=tokenId)
        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/tokens/{tokenId}", "DELETE", **kwargs)

    def verifyAccount(
        self,
        accountId: Optional[str] = None,
        token: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Verify account data.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/verifier

        Args:
            accountId: account id
            token: account JWT token
            login: account login
            password: account password
        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` 204 status code will be returned.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from uuid import uuid4
        >>> import json
        >>>
        >>> response = client.verifyAccount(accountId=client.accountId)
        >>> print(response.statusCode)
        200
        >>> print(response.json)
        {'account_type': 'advanced_user'}

        """
        data = {}
        if login is not None:
            data["login"] = login
        if password is not None:
            data["password"] = password
        if accountId is not None:
            data["account_id"] = accountId
        if token is not None:
            data["token"] = token
        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/credentials_verifier", "POST", body=payload, **kwargs)
