"""
Module realize faces mixin
"""
from typing import Awaitable, List, Optional, Union

from luna3.common.http_objs import RawDescriptor
from luna3.common.luna_response import LunaResponse
from luna3.common.requests import RequestPayload
from luna3.common.utils import getPreparedDescriptors
from luna3.lunavl.mixins.base import MixinStub


class FacesMixin(MixinStub):
    """Faces Mixin to work with faces."""

    def createFace(
        self,
        externalId: Optional[str] = None,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        descriptorSamples: List[str] = None,
        basicAttributesSamples: List[str] = None,
        userData: Optional[str] = None,
        avatar: Optional[str] = None,
        lists: Optional[List[str]] = None,
        eventId: Optional[str] = None,
        contentType: str = "application/json",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Create new face tests!

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/createFace

        Args:
            externalId: user-defined external id for face
            attributeId: an attribute id associated with the face
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.
            userData: user data
            avatar: avatar for face, references to a sample or a user-defined reference
            lists: luna lists for linking face with them
            eventId: event id associated with creating the face
            contentType: content type to use (supported now: "application/json" (default), "application/msgpack")

        Returns:
            LunaResponse: body contains json with info about created face.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> response = client.createFace(
        ...     externalId="2xQ2gprbMePw1s9gw9fvA==",
        ...     userData="test_data05",
        ...     avatar="https://upload.wikimedia.org/wikipedia/commons/thumb/3/3d/Abbey_Lee.jpg/400px-Abbey_Lee.jpg",
        ...     raiseError=True,
        ... )
        >>> print(response.statusCode)
        201
        >>> print(response.json) # doctest: +ELLIPSIS
        {'face_id': ..., 'url': ...}

        """
        bodyJson = {
            "external_id": externalId,
            "user_data": userData,
            "avatar": avatar,
            "lists": lists,
            "event_id": eventId,
        }
        bodyJson = {k: v for k, v in bodyJson.items() if v is not None}
        if attributeId is not None:
            bodyJson.setdefault("attribute", {})["attribute_id"] = attributeId
        if descriptors is not None:
            encode = contentType == "application/json"
            bodyJson.setdefault("attribute", {})["face_descriptors"] = getPreparedDescriptors(
                descriptors, encode=encode
            )
            if descriptorSamples is not None:
                bodyJson["attribute"]["face_descriptor_samples"] = descriptorSamples
        if basicAttributes is not None:
            bodyJson.setdefault("attribute", {})["basic_attributes"] = basicAttributes
            if basicAttributesSamples is not None:
                bodyJson["attribute"]["basic_attributes_samples"] = basicAttributesSamples

        if contentType == "application/json":
            payload = RequestPayload.buildJson(body=bodyJson)
        elif contentType == "application/msgpack":
            payload = RequestPayload.buildMsgpack(body=bodyJson)
        else:
            raise ValueError(f'Unsupported content type "{contentType}"')

        return self.makeRequest(f"{self.baseUri}/faces", "POST", body=payload, **kwargs)

    def getFaces(
        self,
        userData: Optional[str] = None,
        eventId: Optional[str] = None,
        listId: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        page: Optional[int] = 1,
        pageSize: Optional[int] = 10,
        externalIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        targets: Optional[List[str]] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Getting faces which satisfy filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFaces

        Args:
            userData: user data, part of user data, case insensitive
            eventId: event id
            listId: list id
            faceIds: face ids
            timeLt: upper bound of face create time
            timeGte: lower bound of face create time
            page: page count, default 1
            pageSize: page size, default 10, unlimited if -1
            externalIds: list of external identifiers of persons, if they have their own mapping in the system
            faceIdGte: upper bound of face id
            faceIdLt: lower bound of face id
            targets: targets to get faces
            accountIdFilter: account id to get faces

        Returns:
            LunaResponse: body contains json with faces info.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(
        ...     externalId="2xQ2gprbMePw1s9gw9fvA==", userData="test_data05",
        ...     avatar="https://st.kp.yandex.net/images/actor_iphone/iphone360_460188.jpg", raiseError=True
        ... ).json["face_id"]
        >>> response = client.getFaces(
        ...     faceIds=[faceId], targets=["external_id", "user_data", "avatar"], raiseError=True
        ... )
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))
        {
            "faces": [
                {
                    "avatar": "https://st.kp.yandex.net/images/actor_iphone/iphone360_460188.jpg",
                    "external_id": "2xQ2gprbMePw1s9gw9fvA==",
                    "user_data": "test_data05"
                }
            ]
        }

        """
        queries = {
            "user_data": userData,
            "event_id": eventId,
            "list_id": listId,
            "face_ids": faceIds,
            "create_time__lt": timeLt,
            "create_time__gte": timeGte,
            "page": page,
            "page_size": pageSize,
            "external_ids": externalIds,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
            "targets": targets,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/faces", "GET", queryParams=queries, **kwargs)

    def deleteFaces(
        self, faceIds: List[str], ignore: Optional[int] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete faces. Optional you can set, that faces must belong to account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/deleteFaces

        Args:
            faceIds: face id list
            ignore: ignore non-existing faces

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.deleteFaces(faceIds=[faceId])
        >>> print(response.statusCode)
        204

        """
        queries = {"ignore": ignore} if ignore is not None else None
        payload = RequestPayload.buildJson(body={"face_ids": faceIds})
        return self.makeRequest(f"{self.baseUri}/faces", "DELETE", queryParams=queries, body=payload, **kwargs)

    def countFaces(
        self,
        timeLt: Optional[str] = None,
        timeGte: Optional[str] = None,
        faceIds: Optional[List[str]] = None,
        externalIds: Optional[List[str]] = None,
        faceIdGte: Optional[str] = None,
        faceIdLt: Optional[str] = None,
        userData: Optional[str] = None,
        listId: Optional[str] = None,
        eventId: Optional[str] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Getting count of faces which satisfy filters

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFacesCount

        Args:
            timeLt: Upper bound for object create_time
            timeGte: Lower bound for object create_time
            faceIds: list of coma separated faces ids
            externalIds: external id of the face, if it has its own mapping in the system
            faceIdGte: lower face id including boundary
            faceIdLt: upper face id excluding boundary
            userData: Find all objects with user_data which "like" of this parameter
            listId: id of list that contains faces
            eventId: event id
            accountIdFilter: account id to count faces

        Returns:
            LunaResponse: body contains json with number.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> response = client.countFaces(timeGte="2018-08-10", userData="fox1991")
        >>> print(response.statusCode)
        200
        >>> print(response.json) # doctest: +ELLIPSIS
        {'faces_count': ...}

        """
        queries = {
            "user_data": userData,
            "event_id": eventId,
            "list_id": listId,
            "face_ids": faceIds,
            "create_time__lt": timeLt,
            "create_time__gte": timeGte,
            "external_ids": externalIds,
            "face_id__gte": faceIdGte,
            "face_id__lt": faceIdLt,
            "account_id": accountIdFilter,
        }
        return self.makeRequest(f"{self.baseUri}/faces/count", "GET", queryParams=queries, **kwargs)

    def getFace(
        self, faceId: str, targets: Optional[List[str]] = None, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face by id

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFace

        Args:
            faceId: face id
            targets: list of face fields to receive instead of full face
            accountIdFilter: account id to get face

        Returns:
             LunaResponse: body contains json with face information.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json["face_id"]
        >>> response = client.getFace(faceId=faceId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "account_id": ...,
            "avatar": "",
            "create_time": ...,
            "event_id": null,
            "external_id": "",
            "face_id": ...,
            "lists": [],
            "user_data": ""
        }

        """
        queries = {"targets": targets, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "GET", queryParams=queries, **kwargs)

    def patchFace(
        self,
        faceId: str,
        externalId: Optional[str] = None,
        userData: Optional[str] = None,
        avatar: Optional[str] = None,
        eventId: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Update fields of face (user_data, external_id, attribute_id)

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/patchFace

        Args:
            faceId: face id
            externalId: user-defined external id for face
            userData: user data
            avatar: avatar for face, references to a sample or a user-defined reference
            eventId: event id associated with creating the face

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.patchFace(faceId=faceId, userData="test_data05_patched")
        >>> print(response.statusCode)
        204

        """
        data = {}
        if userData is not None:
            data["user_data"] = userData
        if eventId is not None:
            data["event_id"] = eventId
        if externalId is not None:
            data["external_id"] = externalId
        if avatar is not None:
            data["avatar"] = avatar

        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "PATCH", body=payload, **kwargs)

    def deleteFace(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete face by id

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/deleteFace

        Args:
            faceId: face id

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.deleteFace(faceId=faceId)
        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "DELETE", **kwargs)

    def checkFace(
        self, faceId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Check a face with id=face_id existence

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/checkFace

        Args:
            faceId: face id
            accountIdFilter: account id to check face

        Returns:
            LunaResponse:  not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.checkFace(faceId=faceId)
        >>> print(response.statusCode)
        200

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}", "HEAD", queryParams=queries, **kwargs)

    def getFaceAttributes(
        self,
        faceId: str,
        targets: Optional[List[str]] = None,
        descriptorVersion: Optional[int] = None,
        accountIdFilter: Optional[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get face attributes

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFaceAttributes

        Args:
            faceId: face id
            targets: targets to get face attributes. Available targets: `create_time`, `face_descriptor`,
                     `basic_attributes`, `face_descriptor_samples`, `basic_attributes_samples`.
            descriptorVersion: descriptor version
            accountIdFilter: account id to get face attributes

        Returns:
            LunaResponse: body contains json with face information.

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json["face_id"]
        >>> response = client.getFaceAttributes(faceId)
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "attributes": {
                "basic_attributes": null,
                "basic_attributes_samples": [],
                "create_time": null,
                "face_descriptor": null,
                "face_descriptor_samples": []
            }
        }

        """
        queries = {"descriptor_version": descriptorVersion, "targets": targets, "account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes", "GET", queryParams=queries, **kwargs)

    def putFaceAttributes(
        self,
        faceId: str,
        attributeId: Optional[str] = None,
        basicAttributes: Optional[dict] = None,
        descriptors: Optional[List[Union[bytes, RawDescriptor]]] = None,
        basicAttributesSamples: List[str] = None,
        descriptorSamples: List[str] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Put face attributes.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/putFaceAttributes

        Args:
            faceId: face id
            attributeId: attribute id
            descriptors: list of descriptors different versions
            descriptorSamples: list of warp image id from which the attribute was extracted.
            basicAttributes: dict with basic attributes (age, gender, ethnicity)
            basicAttributesSamples: list of warp image id from which the basic attribute was extracted.

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json["face_id"]
        >>> basicAttributes = {"age": 29, "gender": 0, "ethnicity": 3}
        >>> response = client.putFaceAttributes(faceId=faceId, basicAttributes=basicAttributes)
        >>> print(response.statusCode)
        204

        """
        data = {}
        if attributeId is not None:
            data["attribute_id"] = attributeId
        if descriptors is not None:
            data["face_descriptors"] = getPreparedDescriptors(descriptors)
            if descriptorSamples is not None:
                data["face_descriptor_samples"] = descriptorSamples
        if basicAttributes is not None:
            data["basic_attributes"] = basicAttributes
            if basicAttributesSamples is not None:
                data["basic_attributes_samples"] = basicAttributesSamples

        payload = RequestPayload.buildJson(body=data)
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes", "PUT", body=payload, **kwargs)

    def deleteFaceAttributes(self, faceId: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Delete face attributes.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/deleteFaceAttributes  # pylint: disable=line-too-long

        Args:
            faceId: face id

        Returns:
            LunaResponse: not contains response body.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.deleteFaceAttributes(faceId=faceId)
        >>> print(response.statusCode)
        204

        """
        return self.makeRequest(f"{self.baseUri}/faces/{faceId}/attributes", "DELETE", **kwargs)

    def getFaceSamples(
        self, faceId: str, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Get samples which are associated with the face attributes (gender, age, ethnicity, descriptor).

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFaceAttributeSamples  # pylint: disable=line-too-long

        Args:
            faceId: face Id
            accountIdFilter: account id to get face samples

        Returns:
            LunaResponse: body contains json with basic attributes.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> faceId = client.createFace(raiseError=True).json['face_id']
        >>> response = client.getFaceSamples(faceId=faceId)
        >>> print(response.statusCode)
        200
        >>> print(response.json) # doctest: +ELLIPSIS
        {'samples': [...]}

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(
            f"{self.baseUri}/faces/{faceId}/attributes/samples", "GET", queryParams=queries, **kwargs
        )

    def getFacesAttributesCount(
        self, accountIdFilter: Optional[str] = None, **kwargs
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Count faces with attributes. Optional you can set, that faces must belong to account.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/getFaceAttributeCount  # pylint: disable=line-too-long

        Args:
            accountIdFilter: account id to count faces with attributes

        Returns:
            LunaResponse: body contains json with number of faces.

        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>>
        >>> response = client.getFacesAttributesCount()
        >>> print(response.statusCode)
        200
        >>> print(response.json) # doctest: +ELLIPSIS
        {'faces_count': ...}

        """
        queries = {"account_id": accountIdFilter}
        return self.makeRequest(f"{self.baseUri}/faces/attributes/count", "GET", queryParams=queries, **kwargs)
