"""
Module realize liveness mixin
"""

import io
from typing import Awaitable, Dict, List, Optional, Union

import aiohttp
import ujson

from luna3.common.http_objs import BinaryImage
from luna3.common.luna_response import LunaResponse
from luna3.lunavl.mixins.base import MixinStub


def _buildImagesPayload(
    images: List[BinaryImage],
    boundaryString: str,
    meta: Optional[Dict] = None,
    xMeta: Optional[Dict] = None,
) -> aiohttp.MultipartWriter:
    """
    Build payload with images.

    Args:
        images: Single image or images list (supported formats "jpeg, png, tif, ppm, bmp").
        boundaryString: multipart boundary string
        meta: A dict with device information
            Fields:
                os: A str, device os. Required
                manufacture: A str, manufacture
                model: A str, model
        xMeta: A dict with any additional information
    """
    if isinstance(images, BinaryImage):
        images = [images]

    with aiohttp.MultipartWriter("form-data", boundaryString) as writer:
        for img in images:
            newImg = img.clone()
            # Wrap with BytesIO to make aiohttp happy
            imagePart = writer.append(io.BytesIO(newImg.body), newImg.httpMultipartHeaders)
            imagePart.set_content_disposition("form-data", name="image", filename=img.filename, quote_fields=False)

        if meta is not None:
            metaPayload = aiohttp.JsonPayload(meta, dumps=ujson.dumps)
            metaPayload.set_content_disposition("form-data", name="meta")
            writer.append_payload(metaPayload)

        if xMeta is not None:
            xMetaPayload = aiohttp.JsonPayload(xMeta, dumps=ujson.dumps)
            xMetaPayload.set_content_disposition("form-data", name="x_meta")
            writer.append_payload(xMetaPayload)

    return writer


def _buildUrlsPayload(
    urls: List[str],
    meta: Optional[Dict] = None,
    xMeta: Optional[Dict] = None,
) -> aiohttp.JsonPayload:
    """
    Build payload with urls.

    Args:
        urls: list of urls with images
        meta: A dict with device information
            Fields:
                os: A str, device os. Required
                manufacture: A str, manufacture
                model: A str, model
        xMeta: A dict with any additional information
    """
    value = {"urls": urls}
    if meta:
        value["meta"] = meta
    if xMeta:
        value["xMeta"] = xMeta

    return aiohttp.JsonPayload(value)


class LivenessMixin(MixinStub):
    """Methods to work with liveness API"""

    def predictLiveness(
        self,
        images: Optional[Union[BinaryImage, List[BinaryImage]]] = None,
        urls: Optional[List[str]] = None,
        meta: Optional[Dict] = None,
        xMeta: Optional[Dict] = None,
        aggregate: Optional[int] = None,
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Liveness API allows to predict liveness for given images.

        See https://docs.visionlabs.ai/luna/latest/ReferenceManuals/APIReferenceManual.html#operation/predictLiveness

        Args:
            images: Single image or images list (supported formats "jpeg, png, tif, ppm, bmp").
            urls: list of urls with images
            meta: A dict with device information
                Fields:
                    os: A str, device os. Required
                    manufacture: A str, manufacture
                    model: A str, model
            xMeta: A dict with any additional information
            aggregate: Available values : 0, 1; whether to aggregate liveness estimations

        Returns:
            LunaResponse: body contains json with results of predictions (failed and succeeded predictions)

        >>> import json
        >>> from luna3.lunavl.mixins.doctests.config import TEST_CLIENT as client
        >>> from testData.resources import FACE_WARP
        >>>
        >>> response = client.predictLiveness(images=[FACE_WARP])
        >>> print(response.statusCode)
        200
        >>> print(json.dumps(response.json, sort_keys=True, indent=4))  # doctest: +ELLIPSIS
        {
            "images": [
                {
                    "error": {
                        "desc": "Success",
                        "detail": "Success",
                        "error_code": 0,
                        "link": "https://docs.visionlabs.ai/info/luna/troubleshooting/errors-description/code-0"
                    },
                    "filename": "Claudia_Schiffer_warp.jpg",
                    "liveness": {
                        "estimations": {
                            "probability": ...,
                            "quality": ...
                        },
                        "prediction": ...
                    },
                    "status": 1
                }
            ]
        }

        """
        if images:
            payload = _buildImagesPayload(images, self.multipartBoundary, meta, xMeta)
        elif urls:
            payload = _buildUrlsPayload(urls, meta, xMeta)
        else:
            raise ValueError("Either images or urls are required")

        queries = {"aggregate": aggregate}

        return self.makeRequest(f"{self.baseUri}/liveness", "POST", body=payload, queryParams=queries, **kwargs)
