"""
Module implements client for luna-remote-sdk
"""
from numbers import Real
from typing import Awaitable, List, Optional, Union

import aiohttp

from ..common import functions
from ..common.base_api import BaseHTTPAPI
from ..common.http_objs import BinaryImage, UrlForDetection
from ..common.luna_response import LunaResponse
from ..common.requests import RequestPayload
from ..remote_sdk.http_objs import Estimator


class LunaRemoteSDKApi(BaseHTTPAPI):
    """
    Class for request to luna-remote-sdk
    """

    def __init__(
        self,
        origin: Optional[str] = "http://127.0.0.1:5220",
        api: Optional[int] = 1,
        asyncRequest: Optional[bool] = False,
        lunaRequestId: Optional[str] = None,
        totalTimeout: Real = 60,
        connectTimeout: Optional[Real] = None,
        sockReadTimeout: Optional[Real] = None,
        sockConnectTimeout: Optional[Real] = None,
        *,
        session: aiohttp.ClientSession = None,
        echo: bool = False,
        **kwargs,
    ):
        super().__init__(
            origin=origin,
            api=api,
            asyncRequest=asyncRequest,
            lunaRequestId=lunaRequestId,
            totalTimeout=totalTimeout,
            connectTimeout=connectTimeout,
            sockReadTimeout=sockReadTimeout,
            sockConnectTimeout=sockConnectTimeout,
            session=session,
            echo=echo,
            **kwargs,
        )

    def sdk(
        self,
        inputData: Union[BinaryImage, List[BinaryImage], List[UrlForDetection], List[str]],
        estimatePeopleCount: int = None,
        multifacePolicy: int = 1,
        detectFace: Optional[int] = None,
        detectBody: Optional[int] = None,
        estimateHeadPose: Optional[int] = None,
        estimateLandmarks68: Optional[int] = None,
        estimateLandmarks5: Optional[int] = None,
        estimateQuality: Optional[int] = None,
        estimateGaze: Optional[int] = None,
        estimateEyesAttributes: Optional[int] = None,
        estimateEmotions: Optional[int] = None,
        estimateMouthAttributes: Optional[int] = None,
        estimateMask: Optional[int] = None,
        estimateGlasses: Optional[int] = None,
        estimateLiveness: Optional[int] = None,
        estimateFaceWarp: Optional[int] = None,
        estimateBodyWarp: Optional[int] = None,
        aggregateAttributes: Optional[int] = None,
        estimateBasicAttributes: Optional[int] = None,
        estimateBodyDescriptor: Optional[int] = None,
        estimateFaceDescriptor: Optional[int] = None,
        estimateUpperBody: Optional[int] = None,
        estimateLowerBody: Optional[int] = None,
        estimateBodyBasicAttributes: Optional[int] = None,
        estimateAccessories: Optional[int] = None,
        estimateDeepfake: Optional[int] = None,
        pitchThreshold: Optional[int] = None,
        rollThreshold: Optional[int] = None,
        yawThreshold: Optional[int] = None,
        scoreThreshold: float = None,
        maskStates: Optional[List[int]] = None,
        livenessStates: Optional[List[int]] = None,
        deepfakeStates: Optional[List[int]] = None,
        imageType: Optional[int] = None,
        useExifInfo: Optional[int] = None,
        extractExif: Optional[int] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Detect faces on input images.

        Args:
            inputData: one of:
                        - Single image or images list for extraction (supported formats "jpeg, png, tif, ppm, bmp").
                        - list of urls with images
                        - list of images' ids
            estimatePeopleCount: Available values : 0, 1
            multifacePolicy: multiple face detection policy:
                             0 - multiple face detection not allowed,
                             1 - multiple face detection allowed,
                             2 - get best detection from the image
            detectFace: detect faces
            detectBody: detect human bodies
            estimateHeadPose: Available values : 0, 1
            estimateLandmarks5: Available values : 0, 1
            estimateLandmarks68: Available values : 0, 1
            estimateQuality: Available values : 0, 1
            estimateGaze: Available values : 0, 1
            estimateEyesAttributes: Available values : 0, 1
            estimateMouthAttributes: Available values : 0, 1
            estimateEmotions: Available values : 0, 1
            estimateMask: Available values: 0, 1
            estimateGlasses: Available values: 0, 1
            estimateLiveness: Available values: 0, 1
            estimateDeepfake: Available values: 0, 1
            estimateFaceDescriptor: Available values : 0, 1
            estimateBodyDescriptor: Available values : 0, 1
            estimateBasicAttributes: Available values : 0, 1
            estimateBodyWarp: Available values : 0, 1
            estimateFaceWarp: Available values : 0, 1
            estimateUpperBody: Available values : 0, 1
            estimateLowerBody: Available values : 0, 1
            estimateBodyBasicAttributes: Available values : 0, 1
            estimateAccessories: Available values : 0, 1
            pitchThreshold: maximum deviation pitch angle from 0
            rollThreshold: maximum deviation roll angle from 0
            yawThreshold: maximum deviation yaw angle from 0
            scoreThreshold: descriptor garbage score
            maskStates: Array of integer (filter by mask state) (Items Enum:1 2 3)
            livenessStates: Array of integer (filter by liveness state) (Items Enum:0 1 2)
            deepfakeStates: Array of integer (filter by deepfake state) (Items Enum: 0, 1)
            aggregateAttributes: Available values : 0, 1
            imageType: image type (0 - raw image, 1 - face warped image, 2 - body warped image)
            useExifInfo: whether to use exif info for auto orientation
            extractExif: extract EXIF meta information from images
            accept: expected content type

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with samples, exif, filename for succeeded images and
            errors for failed images will be returned.

        """
        queries = {
            "estimate_people_count": estimatePeopleCount,
            "multiface_policy": multifacePolicy,
            "estimate_head_pose": estimateHeadPose,
            "estimate_landmarks68": estimateLandmarks68,
            "estimate_landmarks5": estimateLandmarks5,
            "estimate_quality": estimateQuality,
            "estimate_gaze": estimateGaze,
            "estimate_eyes_attributes": estimateEyesAttributes,
            "estimate_mouth_attributes": estimateMouthAttributes,
            "estimate_emotions": estimateEmotions,
            "estimate_mask": estimateMask,
            "estimate_glasses": estimateGlasses,
            "estimate_liveness": estimateLiveness,
            "estimate_deepfake": estimateDeepfake,
            "pitch_threshold": pitchThreshold,
            "roll_threshold": rollThreshold,
            "yaw_threshold": yawThreshold,
            "image_type": imageType,
            "estimate_body_descriptor": estimateBodyDescriptor,
            "estimate_body_warp": estimateBodyWarp,
            "estimate_upper_body": estimateUpperBody,
            "estimate_lower_body": estimateLowerBody,
            "estimate_body_basic_attributes": estimateBodyBasicAttributes,
            "estimate_accessories": estimateAccessories,
            "estimate_face_warp": estimateFaceWarp,
            "aggregate_attributes": aggregateAttributes,
            "estimate_face_descriptor": estimateFaceDescriptor,
            "score_threshold": scoreThreshold,
            "estimate_basic_attributes": estimateBasicAttributes,
            "detect_face": detectFace,
            "detect_body": detectBody,
            "mask_states": maskStates,
            "liveness_states": livenessStates,
            "deepfake_states": deepfakeStates,
            "use_exif_info": useExifInfo,
            "extract_exif": extractExif,
        }

        payload = functions.createPayloadWithImages(inputData, {}, self.multipartBoundary)
        if accept:
            payload.applyAcceptHeader(accept)

        return self.makeRequest(f"{self.baseUri}/sdk", "POST", queryParams=queries, body=payload, **kwargs)

    def checkISO(
        self,
        inputData: Union[BinaryImage, List[BinaryImage], List[UrlForDetection], List[str]],
        multifacePolicy: int = 0,
        extractExif: int = 0,
        imageType: Optional[int] = None,
        accept: str = "application/msgpack",
        **kwargs,
    ) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Estimate and check input images by ISO.

        Args:
            inputData: images, urls, or sample ids for estimation
            multifacePolicy: multiple face detection policy (0 - not allowed, 1 - allowed, 2 - get the best)
            extractExif: whether to extract exif from the image
            imageType: image type (0 - raw image, 1 - face warped image)
            accept: expected content type

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with filename, exif, error and estimations for images

        """
        queries = {"multiface_policy": multifacePolicy, "image_type": imageType, "extract_exif": extractExif}

        payload = functions.createPayloadWithImages(inputData, {}, self.multipartBoundary)
        if accept:
            payload.applyAcceptHeader(accept)

        return self.makeRequest(f"{self.baseUri}/iso", "POST", queryParams=queries, body=payload, **kwargs)

    def estimator(self, estimator: Estimator, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Estimate image

        Args:
            estimator: request schema

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with samples, exif, filename for succeeded images and
            errors for failed images will be returned.

        """
        body = estimator.asDict()
        payload = RequestPayload.buildMsgpack(body)
        return self.makeRequest(f"{self.baseUri}/estimator", "POST", body=payload, **kwargs)

    def videosdk(self, url: str, **kwargs) -> Union[Awaitable[LunaResponse], LunaResponse]:
        """
        Process video

        Args:
            url: video url

        Returns:
            class:`~.LunaResponse` or *asyncio coroutine* with *LunaResponse*.
            In body of :class: `~.LunaResponse` json with samples, exif, filename for succeeded images and
            errors for failed images will be returned.

        """
        body = {"url": url}
        payload = RequestPayload.buildJson(body)
        return self.makeRequest(f"{self.baseUri}/videosdk", "POST", body=payload, **kwargs)
